/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2024 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* nxsshd, NX protocol compression and NX extensions to this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.a.r.l.                  */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifdef WIN32

#define ERROR
#define WARNING
#undef  TEST
#undef  DEBUG

#include <securitybaseapi.h>
#include "TokenClient.h"

int WriteTokenMessage(HANDLE pipe, TokenMessage message)
{
  int i;

  int totalSize = strlen(tokenMessageBegin) + strlen(tokenMessageEnd) +
                  sizeof(TokenMessage) + 1;

  char *messageBuffer = (char *)malloc(totalSize);

  ZeroMemory((PVOID) messageBuffer, totalSize);

  #ifdef TEST

  debug("WriteTokenMessage: Allocated write buffer with size %d.\n", totalSize);

  #endif

  char *current = messageBuffer;

  for (i = 0; i < MessageComponents; i++)
  {
    switch (i)
    {
      case BeginMessage:
      {
        #ifdef TEST

        debug("WriteTokenMessage: Writing begin message.\n");

        #endif

        StringSafeCatLen(current, tokenMessageBegin, totalSize);

        current += strlen(tokenMessageBegin);

        continue;
      }
      case Message:
      {
        CopyMemory((VOID *) current, (VOID *) &message, sizeof(TokenMessage));

        current += sizeof(TokenMessage);

        continue;
      }
      case EndMessage:
      {
        strncat(current, tokenMessageEnd, strlen(tokenMessageEnd));

        message.complete = 1;

        continue;
      }
      default:
      {
        #ifdef ERROR

        logit("WriteTokenMessage: ERROR! Unexpected message component.\n");

        #endif

        free(messageBuffer);

        return -1;
      }
    }
  }

  DWORD written = 0;

  DWORD writeResult = WriteFile(pipe, messageBuffer, 
                                    totalSize, 
                                        &written, NULL);

  #ifdef TEST

  debug("WriteTokenMessage: Written %lu bytes.\n", written);

  #endif

  if (writeResult == 0)
  {
    #ifdef ERROR

    logit("WriteTokenMessage: ERROR! Writing to client failed.\n");

    logit("WriteTokenMessage: ERROR! Error code '%lu'.\n", GetLastError());

    #endif

    free(messageBuffer);

    return -1;
  }

  free(messageBuffer);

  return 0;
}

static int EnablePrivilege(HANDLE hToken, const char* name)
{
  TOKEN_PRIVILEGES tp;

  ZeroMemory(&tp, sizeof(tp));

  if (LookupPrivilegeValue(NULL, name,
          &tp.Privileges[0].Luid) == 0)
  {
    #ifdef ERROR

    logit("EnablePrivilege: ERROR! LookupPrivilegeValue(). Error [%ld]\n", GetLastError());

    #endif

    return 0;
  }

  tp.PrivilegeCount = 1;
  tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

  if (AdjustTokenPrivileges(hToken, FALSE, &tp,
          sizeof(TOKEN_PRIVILEGES), NULL, 0) == 0)
  {
    #ifdef ERROR

    logit("EnablePrivilege: ERROR! AdjustTokenPrivileges(). Error [%ld].\n", GetLastError());

    #endif

    return 0;
  }

  return 1;
}

static const char *Pipen = "\\\\.\\pipe\\nxserver-WqJhjF9jm0-l0dmS-Hs8dXKc0";

int SendToken(const char *userSid, HANDLE token)
{
  //
  // Open pipe connection before impersonation. Token server pipe 
  // ACL allows 'nx' and 'SYSTEM' users only.
  // 

  #ifdef DEBUG

  debug("SendToken: Opening connection to token server pipe.\n");

  #endif

  HANDLE tokenServerPipe = CreateFile(Pipen, GENERIC_READ | GENERIC_WRITE, 
                                          FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                                              OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
                                                  NULL);

  if (tokenServerPipe == INVALID_HANDLE_VALUE)
  {
    #ifdef ERROR

    logit("SendToken: ERROR! Failed to open token server pipe.\n");

    logit("SendToken: ERROR! Error is '%lu'.\n", GetLastError());

    #endif

    return -1;
  }

  EnablePrivilege(token, SE_TAKE_OWNERSHIP_NAME);

  if (ImpersonateLoggedOnUser(token) == 0)
  {
    #ifdef ERROR

    logit("SendToken: ERROR! Failed to impersonate logged on user.\n");

    logit("SendToken: ERROR! Error is '%lu'.\n", GetLastError());

    #endif

    return -1;
  }

  TokenMessage setTokenMessage;

  StringSafeCopyLen(setTokenMessage.sid, userSid, maxs);

  setTokenMessage.setToken = 1;

  //
  // Write message to token server.
  //

  if (WriteTokenMessage(tokenServerPipe, setTokenMessage) == -1)
  {
    #ifdef ERROR

    logit("SendToken: ERROR! Failed to write message to token server.\n");

    #endif

    RevertToSelf();

    CloseHandle(tokenServerPipe);

    return -1;
  }

  #ifdef DEBUG

  debug("SendToken: Successfully send message to token server.\n");

  #endif

  RevertToSelf();

  CloseHandle(tokenServerPipe);

  return 0;
}

#endif
